#! perl
#
# Yuichi Koike ($Id: dtd2xsd.pl,v 1.14 2000/08/11 16:35:09 connolly Exp $)
# derived from
#
# Dan Connolly
# derived from
#
# Bert Bos <bert@w3.org>
# Created: 17 Mar 1998
#

use strict;

# Alias dictionary
my %alias_dic;
$alias_dic{"URI"} = "uriReference";
$alias_dic{"NUMBER"} = "nonNegativeInteger";

# Handling command line argument
my $targetNS = "http://www.w3.org/namespace/";
my $prefix = "t";
my $alias = 0;
my $file = "";
my %SimpleTypes;
my @AttrGroupPatterns;
my @ModelGroupPatterns;

my %Mixed;
my %ModelGroup;

while ($#ARGV >= 0) {
	my $para = shift(@ARGV);
	if ($para eq "-ns") {
		$targetNS = shift(@ARGV);
	} elsif ($para eq "-prefix") {
		$prefix = shift(@ARGV);
	} elsif ($para eq "-alias") {
		$alias = 1;
	} elsif ($para eq "-simpletype") {
	    my($pat) = shift(@ARGV);
	    my($b) = shift(@ARGV);
	    $SimpleTypes{$pat} = $b;
	} elsif ($para eq "-attrgroup") {
	    push(@AttrGroupPatterns, shift(@ARGV));
	} elsif ($para eq "-modelgroup") {
	    push(@ModelGroupPatterns, shift(@ARGV));
	} else {
		$file = $para;
	}
}

# Variable declaration
my $linelen = 72;

my $PROG = substr($0, rindex($0, "/") + 1);
my $USAGE = "Usage: $PROG file\n";

my $str = "(?:\"([^\"]*)\"|\'([^\']*)\')";
my %pent;				# Parameter entities
my %attributes;			# Attribute lists
my @element;			# Elements in source order
my %model;				# Content models

# Main
$/ = undef;

# Open file, remove comment and include external entity
my $buf = openFile($file);

# Alias treatment
my $alias_ident = "_alias_";
if ($alias eq 1) {
	foreach my $key (keys(%alias_dic)) {
		my $aliaskey = sprintf("%s%s%s", $alias_ident, $key, $alias_ident);
		$buf =~ s/\%$key;/$aliaskey/gsie;
	}
}


# store all parameter entities
while ($buf =~ s/<!ENTITY\s+%\s+(\S+)\s+$str\s*>//sie) {
    my($n, $repltext) = ($1, $2.$3);
    my ($pat);

    next if $pent{$n}; # only the first declaration of an entity counts

    foreach $pat (keys %SimpleTypes){
	if($n =~ /^$pat$/){
	    $buf .= " <!_DATATYPE $n $SimpleTypes{$pat} $repltext> ";
	    $pent{$n} = "#DATATYPEREF $n";
	    undef $n;
	    last;
	}

    }

    foreach $pat (@AttrGroupPatterns){
	if($n =~ /^$pat$/){
	    $buf .= " <!_ATTRGROUP $n $repltext> ";
	    $pent{$n} = "#ATTRGROUPREF $n";
	    undef $n;
	    last;
	}

    }

    foreach $pat (@ModelGroupPatterns){
	if($n =~ /^$pat$/){
	    $buf .= " <!_MODELGROUP $n $repltext> ";
	    $pent{$n} = "#MODELGROUPREF $n";
	    undef $n;
	    last;
	}

    }

    $pent{$n}=$repltext if $n;
}

# remove all general entities
$buf =~ s/<!ENTITY\s+.*?>//gsie;

# loop until parameter entities fully expanded
my $i;
do {
	# count # of substitutions
	$i = 0;
	# expand parameter entities
	$buf =~ s/%([a-zA-Z0-9_\.-]+);?/$i++,$pent{$1}/gse;
} while ($i != 0);

# treat conditional sections
while($buf =~ s/<!\[\s*?INCLUDE\s*?\[(.*)\]\]>/\1/gsie) {};
while($buf =~ s/<!\[\s*?IGNORE\s*?\[.*\]\]>//gsie) {};

# store attribute lists
$buf =~ s/<!ATTLIST\s+(\S+)\s+(.*?)>/store_att($1, $2)/gsie;

# store content models
$buf =~ s/<!ELEMENT\s+(\S+)\s+(.+?)>/store_elt($1, $2)/gsie;

#print "<?xml version='1.0'?>\n";
print "<schema
  xmlns='http://www.w3.org/1999/XMLSchema'
  targetNamespace='$targetNS'
  xmlns:$prefix='$targetNS'>\n";

# find maximum length of non-terminals
#my $maxlen = max(map(length, @element)) + 4;

# write simple type declarations
$buf =~ s/<!_DATATYPE\s+(\S+)\s+(\S+)\s+(.+?)>/write_simpleType($1, $2, $3)/gsie;

# write attribute groups
$buf =~ s/<!_ATTRGROUP\s+(\S+)\s+(.+?)>/write_attrGroup($1, $2)/gsie;

# write model groups
$buf =~ s/<!_MODELGROUP\s+(\S+)\s+(.+?)>/write_modelGroup($1, $2)/gsie;


my($e);

# loop over elements, writing XML schema
foreach $e (@element) {
	my $h = $model{$e};
	my $h2 = $attributes{$e};
	my @model = @$h;

	# print rule for element $e
	print "\n <element name='$e'>\n";

	# print rule for $e's content model
	print "  <complexType ";
	if ($model[0] eq 'EMPTY') {
		if (! $h2 ) {
			print "content='empty'/>\n";
		} else {
			print "content='empty'>\n";
		}
	} else {
	    print STDERR "==mixed? @model\n"; #@@
		if (&isMixed(@model)) {
			print "content='mixed'>\n";
		}else{
			print "content='elementOnly'>\n";
		}

		my @list = &makeChildList('', @model);
		&printChildList(3, @list);
	}


	# print rule for $e's attributes
	if (! $h2) {
	    # nothing
	} else {
	    &printAttrDecls(@$h2);
	}


	if ($h2 || $model[0] ne 'EMPTY') {
		print "  </complexType>\n";
	}
	print " </element>\n";
}

print "</schema>\n";
exit;

sub printSpace
{
	my ($num) = $_[0];
	for (my $i=0; $i<$num; $i++) {
		print " ";
	}
}

sub printChildList
{
	my ($num, @list) = @_;

	my @currentTag = ();
	for (my $i=0; $i<= $#list; $i++) {
		my $n = $list[$i];

		if ($n eq 0 || $n eq 1 || $n eq 2 || $n eq 3) {
			if ($currentTag[$#currentTag] eq "" && $n eq 0) {
				push(@currentTag, "");
			} else {
				printSpace($num); $num++;
				print "<sequence";
				if ($n eq 1) {
					print " minOccurs='0' maxOccurs='unbounded'";
				} elsif ($n eq 2) {
					print " maxOccurs='unbounded'";
				} elsif ($n eq 3) {
					print " minOccurs='0' maxOccurs='1'";
				}
				print ">\n";
				push(@currentTag, "sequence");
			}
		} elsif ($n eq 10 || $n eq 11 || $n eq 12 || $n eq 13) {
			printSpace($num); $num++;
			print "<choice";
			if ($n eq 11) {
				print " minOccurs='0' maxOccurs='unbounded'";
			} elsif ($n eq 12) {
				print " maxOccurs='unbounded'";
			} elsif ($n eq 13) {
				print " minOccurs='0' maxOccurs='1'";
			}
			print ">\n";
			push(@currentTag, "choice");
		} elsif ($n eq 20) {
			my $tag = pop(@currentTag);
			if ($tag ne "") {
				$num--; printSpace($num);
				print "</", $tag, ">\n";
			}
		} else {
			printSpace($num);
			if($n eq '#MODELGROUPREF'){
			    print "<group ref='$prefix:$list[++$i]'";
			}else{
			    print "<element ref='$prefix:$n'";
			}

			if ($currentTag[$#currentTag] ne "choice") {
				if ($list[$i+1] eq "+") {
					print " maxOccurs='unbounded'";
					$i++;
				} elsif ($list[$i+1] eq "?") {
					print " minOccurs='0' maxOccurs='1'";
					$i++;
				} elsif ($list[$i+1] eq "*") {
					print " minOccurs='0' maxOccurs='unbounded'";
					$i++;
				}
			}
				print "/>\n";
		}
	}
}

sub makeChildList {
	my ($groupName, @model) = @_;
	my @ret = ();
	my @brace = ();
	for (my $i=0; $i<=$#model; $i++) {
		my $n = $model[$i];

		if ($n eq "(") {
			push(@ret, 0);
			push(@brace, $#ret);
		} elsif ($n eq ")") {
			if ($model[$i+1] eq "*") {
				$ret[$brace[$#brace]] += 1;
				$i++;
			} elsif ($model[$i+1] eq "+") {
				$ret[$brace[$#brace]] += 2;
				$i++;
			} elsif ($model[$i+1] eq "?") {
				$ret[$brace[$#brace]] += 3;
				$i++;
			}
			pop(@brace);
			push(@ret, 20);
		} elsif ($n eq ",") {
			$ret[$brace[$#brace]] = 0;
		} elsif ($n eq "|") {
			$ret[$brace[$#brace]] = 10;
		} elsif ($n eq "#PCDATA") {
			if ($model[$i+1] eq "|") {
				$i++;
			}
			if($groupName){
			    $Mixed{$groupName} = 1;
			}
		} else {
			push(@ret, $n);
		}
	}


	return @ret;
}


sub printAttrDecls{
    my @atts = @_;

    for (my $i = 0; $i <= $#atts; $i++) {
	if($atts[$i] eq '#ATTRGROUPREF'){
	    print "   <attributeGroup ref='$prefix:$atts[$i+1]'/>\n";
	    $i ++;
	}else{
	    # attribute name
	    print "   <attribute name='$atts[$i]'";

	    # attribute type
	    my @enume;
	    $i++;
	    if ($atts[$i] eq "(") {
		# like `attname ( yes | no ) #REQUIRED`
		$i++;
		while ($atts[$i] ne ")") {
		    if ($atts[$i] ne "|") {
			push(@enume, $atts[$i]);
		    }
		    $i++;
		}
	    } elsif ($atts[$i] eq '#DATATYPEREF'){
		print " type='$prefix:$atts[++$i]'";
	    } elsif ($alias eq 1 && $atts[$i] =~ s/$alias_ident//gsie) {
		# alias special
		print " type='$alias_dic{$atts[$i]}'";
	    } elsif($atts[$i] =~ /ID|IDREF|ENTITY|NOTATION|IDREFS|ENTITIES|NMTOKEN|NMTOKENS/) {
		# common type for DTD and Schema
		print " type='$atts[$i]'";
	    } else {
		# `attname CDATA #REQUIRED`
		print " type='string'";
	    }

	    $i++;

	    # #FIXED
	    if($atts[$i] eq "#FIXED") {
		$i++;
		print " use='fixed' value='$atts[$i]'/>\n";
	    } else {
		# minOccurs
		if ($atts[$i] eq "#REQUIRED") {
		    print " use='required'";
		} elsif ($atts[$i] eq "#IMPLIED") {
		    print " use='optional'";
		} else {
		    print " use='default' value='$atts[$i]'";
		}

		# enumerate
		if ($#enume eq -1) {
		    print "/>\n";
		} else {
		    print ">\n";
		    print "    <simpleType base='string'>\n";
		    &write_enum(@enume);
		    print "    </simpleType>\n";
		    print "   </attribute>\n";
		}
	    }
	}
    }
}

sub write_enum{
    my(@enume) = @_;

    for (my $j = 0; $j <= $#enume; $j++) {
	print "     <enumeration value='$enume[$j]'/>\n";
    }
}


# Parse a string into an array of "words".
# Words are whitespace-separated sequences of non-whitespace characters,
# or quoted strings ("" or ''), with the quotes removed.
# HACK: added () stuff for attlist stuff
# Parse words for attribute list
sub parsewords {
	my $line = $_[0];
	$line =~ s/(\(|\)|\|)/ $1 /g;
	my @words = ();

	while ($line ne '') {
		if ($line =~ /^\s+/) {
			# Skip whitespace
		} elsif ($line =~ /^\"((?:[^\"]|\\\")*)\"/) {
			push(@words, $1);
		} elsif ($line =~ /^\'((?:[^\']|\\\')*)\'/) {
			push(@words, $1);
		} elsif ($line =~ /^\S+/) {
			push(@words, $&);
		} else {
			die "Cannot happen\n";
		}
		$line = $';
	}
    return @words;
}

# Store content model, return empty string
sub store_elt
{
	my ($name, $model) = @_;
	$model =~ s/\s+/ /gso;
	push(@element, $name);

	my @words;
	while ($model =~ s/^\s*(\(|\)|,|\+|\?|\||[\w_\.-]+|\#\w+|\*)//) {
		push(@words, $1);
	};
	$model{$name} = [ @words ];
	return '';
}


# Store attribute list, return empty string
sub store_att
{
	my ($element, $atts) = @_;
	my @words = parsewords($atts);
	$attributes{$element} = [ @words ];
	return '';
}

sub write_simpleType{
    my($n, $b, $stuff) = @_;
    my @words = parsewords($stuff);

    print "\n  <simpleType name='$n' base='$b'>\n";
#    print STDERR "\n==stuff:\n$stuff \n\n===\n", join('|', @words);

    my $i = 0;
    my @enume;

    if ($words[$i] eq "(") {
	$i++;
	while ($words[$i] ne ")") {
	    if ($words[$i] ne "|") {
		push(@enume, $words[$i]);
	    }
	    $i++;
	}
	write_enum(@enume);
    }

    print "  </simpleType>\n";
}

sub write_attrGroup{
    my($n, $stuff) = @_;
    my @words = parsewords($stuff);

    print "\n  <attributeGroup name='$n'>\n";
#    print STDERR "\n==stuff:\n$stuff \n\n===\n", join('|', @words);
    printAttrDecls(@words);
    print "  </attributeGroup>\n";
}

sub write_modelGroup{
    my($n, $stuff) = @_;
    my @words = parsewords($stuff);

    print "\n  <group name='$n'>\n";
#    print "<!-- $stuff -->\n";

    my @list = &makeChildList($n, '(', @words, ')');
    &printChildList(3, @list);

    $ModelGroup{$n} = \@list;

    print "  </group>\n";
}

sub isMixed{
    my(@model) = @_;

    my($i);

    print STDERR "++ mixed? @model\n"; #@@

    for($i = 0; $i <= $#model; $i++){
	if ($model[$i] eq '#PCDATA'
	    || ($model[$i] eq '#MODELGROUPREF'
		&& $Mixed{$model[$i+1]}) ){
	    print STDERR "++ yes! $i @model\n"; #@@
	    return 1;
	}
    }

    print STDERR "++ no. @model\n"; #@@

    return 0;
}

# Return maximum value of an array of numbers
sub max
{
	my $max = $_[0];
	foreach my $i (@_) {
		if ($i > $max) {$max = $i;}
	}
	return $max;
}


# 1) Open file
# 2) Remove comment, processing instructions, and general entities
# 3) Include external parameter entities recursively
# 4) Return the contents of opened file
sub openFile {
	my $file = $_[0];

	my %extent;
	my $bufbuf;
	if ($file ne "") {
		print STDERR "open $file ";
		if(! open AAA, $file) {
			print STDERR " failed!!\n";
			return "";
		}
		print STDERR " successful\n";
		$bufbuf = <AAA>;
	} else {
		print STDERR "open STDIN successful\n";
		$bufbuf = <>;
	}

	# remove comments
	$bufbuf =~ s/<!--.*?-->//gso;

	# remove processing instructions
	$bufbuf =~ s/<\?.*?>//gso;

	# store external parameter entities
	while ($bufbuf =~ s/<!ENTITY\s+%\s+(\S+)\s+PUBLIC\s+$str\s+$str.*?>//sie) {
		$extent{$1} = $4.$5;
	}
	while ($bufbuf =~ s/<!ENTITY\s+%\s+(\S+)\s+SYSTEM\s+$str.*?>//sie) {
		$extent{$1} = $2.$3;
	}

	# read external entity files
	foreach my $key (keys(%extent)) {
		$bufbuf =~ s/%$key;/openFile($extent{$key})/gsie;
	}

	return $bufbuf;
}
